# -*- coding: utf-8 -*-
"""
*****************************************************************************************
 *                                                                                     *
*   Ce programme est un logiciel libre ; vous pouvez le redistribuer et/ou le modifier  *
*   selon les termes de la Licence Publique Générale GNU telle que publiée par          *
*   la Free Software Foundation ; soit la version 2 de la Licence, ou                   *
*   (à votre choix) toute version ultérieure.                                           *
 *                                                                                     *
*****************************************************************************************
"""

import os, shutil
import tempfile
import xml.etree.ElementTree as ET
from qgis.core import *
from qgis.PyQt.QtCore import *
from qgis.PyQt.QtWidgets import *
from qgis.PyQt import uic

from .ModeleListeCouches import ModeleListeCouches
from . import resources


# Initialize Qt resources from file resources.py
def replaceText(node, newText):
    if node.firstChild.nodeType != node.TEXT_NODE:
        raise Exception("node does not contain text")

    node.firstChild.replaceWholeText(newText)


from .resources import *

FORM_CLASS, _ = uic.loadUiType(os.path.join(
    os.path.dirname(__file__), 'QPackage_dialog_base.ui'))

class LayerInfo:
    def __init__(self, crs_origin):
        self.crs_origin = crs_origin

class QPackageDialog(QDialog, FORM_CLASS):

    copierCouchesTerminee = pyqtSignal()  # Signal pour indiquer la fin de copierCouches

    def __init__(self, iface, parent=None):

        self.gpkg = None
        self.list_layers = []
        self.list_gpkg_layers = []
        self.gpkg_layer = None
        self.dirNameQgs = None
        self.dirNameQgz = None
        self.dirNameGpkg = None
        self.iface = iface
        """Constructor."""
        super(QPackageDialog, self).__init__(parent)
        # self.crs_origin = None  # Initialise crs_origin à None
        # uic.loadUi(os.path.join(os.path.dirname(__file__), 'QPackage_dialog_base.ui'), self)
        ui_path = os.path.join(os.path.dirname(__file__), 'QPackage_dialog_base.ui')
        uic.loadUi(ui_path, self)
        # self.setupUi(self)

        self.transform_context = QgsCoordinateTransformContext()


    def tr(self, message):
        return QCoreApplication.translate('qgz2gpkgDialog', message)


    def chercherRepertoire(self):
        filename = QFileDialog.getExistingDirectory(
            self,
            caption=self.tr("Select directory to Fling from..."),
            directory=QDir.currentPath()
        )
        if filename:
            self._repertoire.setText(filename)

    def chargerCouches(self):
        # Récupérer le CRS du projet
        crs = QgsProject.instance().crs()
        self.crs_origin = crs.authid()  # Définit crs_origin sur l'instance
        print ('self.crs_origin : ' + self.crs_origin)
    # self._listeprojections.clear()
        data = []
        for layer in QgsProject.instance().mapLayers().values():
            casecocher = QCheckBox(layer.name())
            if layer.type() == QgsMapLayer.VectorLayer:
                casecocher.setChecked(True)
            data.append(casecocher)
        self._tableau.setModel(ModeleListeCouches(data))
        # Réinitialiser la barre de progression après l'opération
        self._progression.setRange(0, 100)
        self._progression.setValue(0)


    def on_copy_clicked(self):
        """Méthode appelée lorsque le bouton _copy est cliqué."""
        # Démarre l'animation de la barre de progression
        self._progression.setRange(0, 0)  # Met la QProgressBar en mode indéterminé

        # Lance la copie des couches
        self.copierCouches()

        # Réinitialise la barre de progression après l'opération
        self._progression.setRange(100)

    def copierCouches(self, nbrecouches=None):
        self.end_copy = 0
        QgsProject.instance().write()

        model = self._tableau.model()
        root = QgsProject.instance().layerTreeRoot()

        self._progression.setValue(0)
        nbrecouches = sum(1 for row in model.getDonnees() if row.isChecked())
        pas = 100 / nbrecouches if nbrecouches > 0 else 0
        progression = 0

        if not self._repertoire.text():
            QMessageBox.critical(self, self.tr("QPackage"), self.tr("You must choose the destination directory"),
                                 QMessageBox.Ok)
            return  # Arrêter l'exécution si le répertoire n'est pas choisi

        self.new_project = QgsProject()
        self.new_project.clear()

        base_project = QgsProject.instance().fileName()
        base_project_name = os.path.splitext(os.path.basename(base_project))[0]
        self.base_project_path = os.path.dirname(base_project)
        self.projet_path = os.path.join(self._repertoire.text(), f"{base_project_name}.qgz")
        self.new_project.setFileName(self.projet_path)

        crs_base = QgsProject.instance().crs()
        self.new_project.setCrs(crs_base)

        QgsMessageLog.logMessage(self.tr("Starting layers copy"), level=Qgis.Info)

        ordered_layers = root.children()[::-1]

        for node in ordered_layers:
            if not isinstance(node, QgsLayerTreeLayer):
                continue

            layer = node.layer()
            if layer is None:
                QgsMessageLog.logMessage(self.tr(f"Layer with name {node.name()} not found"), level=Qgis.Warning)
                continue

            QgsMessageLog.logMessage(self.tr(f"Processing layer {layer.name()}"), level=Qgis.Info)
            destination_path = os.path.join(self._repertoire.text(), layer.name())

            if layer.type() == QgsMapLayer.VectorLayer:
                self.copy_vector_layer(layer, destination_path, crs_base)

            elif layer.type() == QgsMapLayer.RasterLayer:
                self.copy_raster_layer(layer, destination_path)

            progression += pas
            self._progression.setValue(progression)

        # Ajout des nouvelles routines ici
        extensions_fichiers = ('.pdf', '.csv', '.jpg', '.png', '.xls', '.xlsx', '.ods')
        # Utilisez le répertoire source correct
        source_directory = self.base_project_path
        fichiers_a_copier = self.rechercher_fichiers_avec_extensions(source_directory, extensions_fichiers)
        print('fichiers à copier : ' + str(fichiers_a_copier))
        self.copier_fichiers(fichiers_a_copier, self._repertoire.text())

        noms_sous_dossiers = ['photos', 'Photos', 'photo', 'Photo']
        sous_dossiers_a_copier = self.rechercher_sous_dossiers(source_directory, noms_sous_dossiers)
        print('sous_dossiers_a_copier : ' + str(sous_dossiers_a_copier))
        self.copier_sous_dossiers(sous_dossiers_a_copier, self._repertoire.text())

        QgsMessageLog.logMessage(self.tr("Updating SVG, UI, PY paths"), level=Qgis.Info)
        self.update_layer_paths()

        QgsMessageLog.logMessage(self.tr(f"Saving project {self.projet_path}"), level=Qgis.Info)
        self.new_project.write(self.projet_path)

        self._progression.setValue(100)
        QgsMessageLog.logMessage(self.tr("Operation completed successfully"), level=Qgis.Info)

        self.copierCouchesTerminee.emit()

    def copy_vector_layer(self, layer, destination_path, crs_base):
        """Copier une couche vectorielle vers le répertoire de destination."""
        base_path = destination_path + ".shp"
        # Test existence du dossier
        if not os.path.exists(layer.source()):
            self.show_warning_popup(layer.source())
            return

        QgsVectorFileWriter.writeAsVectorFormat(
            layer,
            base_path,
            "utf-8",
            crs_base,
            "ESRI Shapefile"
        )
        new_layer = QgsVectorLayer(base_path, layer.name(), "ogr")
        if new_layer.isValid():
            self.apply_style_from_source(layer, new_layer)
            self.new_project.addMapLayer(new_layer)
        else:
            QgsMessageLog.logMessage(self.tr(f"Vector layer {layer.name()} is not valid"), level=Qgis.Warning)

    def copy_raster_layer(self, layer, destination_path):
        """Copier une couche raster vers le répertoire de destination."""
        raster_path = layer.publicSource()
        # Test existence du dossier
        if not os.path.exists(raster_path):
            self.show_warning_popup(raster_path)
            return

        file_size = os.path.getsize(raster_path)

        destination_full_path = os.path.normpath(
            os.path.join(self._repertoire.text(), layer.name() + os.path.splitext(raster_path)[1]))
        if os.path.normpath(raster_path) == destination_full_path:
            QgsMessageLog.logMessage(self.tr(f"Source and destination for {layer.name()} are the same, skipping copy."),
                                     level=Qgis.Warning)
        else:
            if file_size > 1e9:
                QMessageBox.warning(self, self.tr("QPackage"),
                                    self.tr(
                                        f"The file {os.path.basename(raster_path)} is large and may take a long time to copy."))

            shutil.copy2(raster_path, destination_full_path)

            new_layer = QgsRasterLayer(destination_full_path, layer.name())
            if new_layer.isValid():
                self.new_project.addMapLayer(new_layer)
            else:
                QgsMessageLog.logMessage(self.tr(f"Raster layer {layer.name()} is not valid"), level=Qgis.Warning)

    def update_layer_paths(self):
        """Mettre à jour les chemins des fichiers SVG, UI, et PY dans les propriétés des couches."""
        for layer in self.new_project.mapLayers().values():
            if isinstance(layer, QgsVectorLayer):
                self.update_svg_paths(layer, self._repertoire.text())
                self.update_ui_py_paths(layer, self._repertoire.text())

    def apply_style_from_source(self, source_layer, new_layer):
        """Appliquer le style d'une couche source à une nouvelle couche, en mettant à jour les chemins des fichiers SVG, UI, et PY."""
        # Enregistrer le style de la couche source dans un fichier temporaire
        with tempfile.NamedTemporaryFile(delete=False, suffix='.qml') as temp_file:
            source_layer.saveNamedStyle(temp_file.name)
            temp_file_path = temp_file.name

        # Copier les fichiers SVG, UI, et PY vers le répertoire approprié et mettre à jour les chemins dans le fichier QML
        self.copy_svg_files(temp_file_path, self._repertoire.text())
        self.copy_ui_py_files(temp_file_path, self._repertoire.text())

        # Appliquer le style mis à jour à la nouvelle couche
        updated_qml_path = temp_file_path.replace('.qml', '_updated.qml')
        if new_layer.loadNamedStyle(updated_qml_path):
            QgsMessageLog.logMessage(self.tr(f"Style successfully applied to {new_layer.name()}"), level=Qgis.Info)
        else:
            QgsMessageLog.logMessage(self.tr(f"Failed to apply style to {new_layer.name()}"), level=Qgis.Warning)

        # Nettoyer les fichiers temporaires
        os.remove(temp_file_path)
        os.remove(updated_qml_path)

    def copy_svg_files(self, qml_path, new_project_dir):
        """Copier les fichiers SVG d'un fichier QML de style vers un sous-dossier 'symbols' dans le répertoire du nouveau projet."""
        tree = ET.parse(qml_path)
        root = tree.getroot()

        # Créer le sous-dossier symbols
        symbols_dir = os.path.join(new_project_dir, 'symbols')
        os.makedirs(symbols_dir, exist_ok=True)

        # Trouver les balises SVG dans le fichier QML
        svg_files = root.findall(".//svg")
        for svg in svg_files:
            svg_path = svg.attrib.get('file')
            if svg_path and os.path.exists(svg_path):
                # Copier le fichier SVG vers le sous-dossier symbols
                svg_file_name = os.path.basename(svg_path)
                new_svg_path = os.path.join(symbols_dir, svg_file_name)
                shutil.copy2(svg_path, new_svg_path)

                # Mettre à jour le chemin du fichier dans le fichier QML
                svg.set('file', new_svg_path.replace(os.path.sep, '/'))

        # Enregistrer le fichier QML mis à jour
        updated_qml_path = qml_path.replace('.qml', '_updated.qml')
        tree.write(updated_qml_path)

    def copy_ui_py_files(self, qml_path, new_project_dir):
        """Copier les fichiers UI et PY d'un fichier QML de style vers les sous-dossiers 'forms' et 'pyfiles' dans le répertoire du nouveau projet."""
        tree = ET.parse(qml_path)
        root = tree.getroot()

        # Créer les sous-dossiers forms et pyfiles
        forms_dir = os.path.join(new_project_dir, 'forms')
        pyfiles_dir = os.path.join(new_project_dir, 'pyfiles')
        os.makedirs(forms_dir, exist_ok=True)
        os.makedirs(pyfiles_dir, exist_ok=True)

        # Trouver les balises contenant les chemins des fichiers UI et PY
        ui_files = root.findall(".//editform")
        py_files = root.findall(".//editforminitfilepath")

        # Copier les fichiers UI vers le sous-dossier forms
        for ui in ui_files:
            ui_path = ui.text
            if ui_path and os.path.exists(ui_path):
                ui_file_name = os.path.basename(ui_path)
                new_ui_path = os.path.join(forms_dir, ui_file_name)
                shutil.copy2(ui_path, new_ui_path)
                ui.text = new_ui_path.replace(os.path.sep, '/')
            elif ui_path:
                self.show_warning_popup(ui_path)

        # Copier les fichiers PY vers le sous-dossier pyfiles
        for py in py_files:
            py_path = py.text
            if py_path and os.path.exists(py_path):
                py_file_name = os.path.basename(py_path)
                new_py_path = os.path.join(pyfiles_dir, py_file_name)
                shutil.copy2(py_path, new_py_path)
                py.text = new_py_path.replace(os.path.sep, '/')
            elif py_path:
                self.show_warning_popup(py_path)

        # Enregistrer le fichier QML mis à jour
        updated_qml_path = qml_path.replace('.qml', '_updated.qml')
        tree.write(updated_qml_path)

    def update_svg_paths(self, layer, new_project_dir):
        """Mettre à jour les chemins des fichiers SVG dans les propriétés des couches."""
        symbols_dir = os.path.join(new_project_dir, 'symbols')
        symbols = layer.renderer().symbols(QgsRenderContext())
        for symbol in symbols:
            for symbol_layer in symbol.symbolLayers():
                if isinstance(symbol_layer, QgsSvgMarkerSymbolLayer):
                    svg_path = symbol_layer.path()
                    if os.path.exists(svg_path):
                        svg_file_name = os.path.basename(svg_path)
                        new_svg_path = os.path.join(symbols_dir, svg_file_name)
                        shutil.copy2(svg_path, new_svg_path)
                        symbol_layer.setPath(new_svg_path.replace(os.path.sep, '/'))
                    else:
                        self.show_warning_popup(svg_path)

    def update_ui_py_paths(self, layer, new_project_dir):
        """Mettre à jour les chemins des fichiers UI et PY dans les propriétés des couches."""
        if layer.type() == QgsMapLayer.VectorLayer:
            # Accéder directement aux propriétés personnalisées de la couche
            ui_path = layer.customProperty('ui_path', '')
            py_path = layer.customProperty('py_path', '')

            # Mettre à jour les chemins des fichiers .ui et .py
            new_ui_path = self.update_path_in_url(ui_path, new_project_dir, '.ui', 'forms')
            new_py_path = self.update_path_in_url(py_path, new_project_dir, '.py', 'pyfiles')

            # Définir de nouvelles valeurs pour les propriétés personnalisées
            layer.setCustomProperty('ui_path', new_ui_path)
            layer.setCustomProperty('py_path', new_py_path)

            QgsMessageLog.logMessage(
                self.tr(f"Updated UI/PY paths for {layer.name()}: UI={new_ui_path}, PY={new_py_path}"),
                level=Qgis.Info
            )

    def update_path_in_url(self, original_path, new_project_dir, extension, subdir):
        """Mettre à jour un chemin de fichier dans une URL vers un sous-dossier spécifique dans le nouveau répertoire."""
        if original_path and os.path.exists(original_path):
            new_path = os.path.join(new_project_dir, subdir, os.path.basename(original_path))
            shutil.copy2(original_path, new_path)
            return new_path
        elif original_path:
            self.show_warning_popup(original_path)
        return ''

    def show_warning_popup(self, missing_path):
        """Afficher une fenêtre d'alerte pour les chemins manquants."""
        message = self.tr(
            f"The file at path '{missing_path}' was not found. Please check the file path or the drive letter.")
        QMessageBox.warning(self, self.tr("QPackage"), message)

    def rechercher_fichiers_avec_extensions(self, dossier_source, extensions):
        """Recherche des fichiers avec des extensions spécifiques dans le dossier source."""
        fichiers_a_copier = []
        # for fichier in os.listdir(dossier_source):
        #     if fichier.lower().endswith(tuple(ext.lower() for ext in extensions)):
        #         fichiers_a_copier.append(fichier)
        for root, _, files in os.walk(dossier_source):
            for file_name in files:
                if file_name.lower().endswith(extensions):
                    fichiers_a_copier.append(os.path.join(root, file_name))
        return fichiers_a_copier

    def copier_fichiers(self, fichiers_a_copier, dossier_cible):
        """Copie les fichiers spécifiés dans le dossier cible."""
        for fichier in fichiers_a_copier:
            destination = os.path.join(dossier_cible, os.path.basename(fichier))
            shutil.copy2(fichier, destination)
            QgsMessageLog.logMessage(self.tr(f"Fichier copié : {fichier} -> {destination}"), level=Qgis.Info)

    def rechercher_sous_dossiers(self, dossier_source, noms_sous_dossiers):
        """Recherche des sous-dossiers spécifiques dans le dossier source."""
        sous_dossiers_a_copier = []
        for sous_dossier in noms_sous_dossiers:
            chemin_sous_dossier = os.path.join(dossier_source, sous_dossier)
            if os.path.isdir(chemin_sous_dossier):
                sous_dossiers_a_copier.append(chemin_sous_dossier)
        return sous_dossiers_a_copier

    def copier_sous_dossiers(self, sous_dossiers, destination):
        """Copie les sous-dossiers dans le répertoire de destination."""
        for sous_dossier in sous_dossiers:
            destination_sous_dossier = os.path.join(destination, os.path.basename(sous_dossier))
            if os.path.exists(destination_sous_dossier):
                shutil.rmtree(destination_sous_dossier)
            shutil.copytree(sous_dossier, destination_sous_dossier)
